import unittest
import os
import pyarrow as pa
from cortex.utils import generate_token
from pprint import pprint

from ds_capability.components.commons import Commons

from ds_capability.handlers.mc_handlers import McSourceHandler, McPersistHandler
from ds_core.handlers.abstract_handlers import ConnectorContract, HandlerFactory


class ManagedContentHandlerTest(unittest.TestCase):

    def setUp(self):
        PAT = {"jwk": {"crv": "Ed25519", "d": "chtqzYc7igE62XMI9U0y_NbOjjDGoO2k6AnNCvasK_4",
                       "x": "oJa4EVeM0rhJGZUs_DH-MuKOsmrWS0Kj3E_TDPY_k-Q", "kty": "OKP",
                       "kid": "5x0yfsfU57eL_BLr9pnaj1fyLyufFJYpza_N8A8JSH8", "alg": "EdDSA"}, "issuer": "tecnotree.com",
               "audience": "sensa", "username": "1652d24e-7974-4a83-896d-0eac230a1c62",
               "url": "https://api.dci-dev.dev-eks.insights.ai"}
        os.environ["TOKEN"] = generate_token(PAT)
        os.environ["API_ENDPOINT"] = "https://api.dci-dev.dev-eks.insights.ai"
        os.environ['PROJECT'] = "bptest"

    def tearDown(self):
        del os.environ["TOKEN"]
        del os.environ["API_ENDPOINT"]
        del os.environ['PROJECT']

    def test_mc_handler(self):
        """
        required params: 
            - uri : used as the key for managed content
            - api_endpoint: cortex api endpoint
            - token: cortex token
            - project: cortex project
        """
        cc = ConnectorContract(uri='mc://test/test.parquet', module_name='', handler='')
        handler = McPersistHandler(cc)
        tbl = pa.table([pa.array([1,2,3,4,5])], names=['a'])
        handler.persist_canonical(tbl)
        result = handler.load_canonical()
        self.assertTrue(handler.has_changed())
        self.assertFalse(handler.has_changed())
        self.assertEqual((5,1),result.shape)
        self.assertTrue(handler.remove_canonical())
        #
        cc = ConnectorContract(uri='mc://test/test.csv', module_name='', handler='')
        handler = McPersistHandler(cc)
        tbl = pa.table([pa.array([1,2,3,4,5])], names=['a'])
        handler.persist_canonical(tbl)
        result = handler.load_canonical()
        self.assertTrue(handler.has_changed())
        self.assertFalse(handler.has_changed())
        self.assertEqual((5,1),result.shape)
        self.assertTrue(handler.remove_canonical())


if __name__ == '__main__':
    unittest.main()


import unittest
import os
from pathlib import Path
import shutil
import ast
from pprint import pprint

import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.parquet as pq
from ds_capability.sample.sample_data import MappedSample

from ds_capability import *
from ds_capability.intent.feature_engineer_intent import FeatureEngineerIntent
from ds_core.properties.property_manager import PropertyManager

# Pandas setup
pd.set_option('max_colwidth', 320)
pd.set_option('display.max_rows', 100)
pd.set_option('display.max_columns', 99)
pd.set_option('expand_frame_repr', True)


class FeatureEngineerTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        # clean out any old environments
        for key in os.environ.keys():
            if key.startswith('HADRON'):
                del os.environ[key]
        # Local Domain Contract
        os.environ['HADRON_PM_PATH'] = os.path.join('working', 'contracts')
        os.environ['HADRON_PM_TYPE'] = 'json'
        # Local Connectivity
        os.environ['HADRON_DEFAULT_PATH'] = Path('working/data').as_posix()
        # Specialist Component
        try:
            os.makedirs(os.environ['HADRON_PM_PATH'])
        except OSError:
            pass
        try:
            os.makedirs(os.environ['HADRON_DEFAULT_PATH'])
        except OSError:
            pass
        PropertyManager._remove_all()

    def tearDown(self):
        try:
            shutil.rmtree('working')
        except OSError:
            pass

    def test_for_smoke(self):
        fe = FeatureEngineer.from_memory()
        tools: FeatureEngineerIntent = fe.tools
        tbl = tools.get_synthetic_data_types(100)
        self.assertEqual((100, 6), tbl.shape)
        self.assertCountEqual(['cat', 'num', 'int', 'bool', 'date', 'string'], tbl.column_names)
        tbl = tools.get_synthetic_data_types(100, extend=True, prob_nulls=0.03)
        self.assertEqual((100, 19), tbl.shape)

    def test_event_event_singleton(self):
        fe = FeatureEngineer.from_env('task1', has_contract=False)
        fe.set_persist_uri("event://task1_outcome")
        tbl = fe.tools.get_synthetic_data_types(size=10)
        # pprint(pm_view('feature_build', 'task1', 'intent'))
        fe.run_component_pipeline()
        h = fe.pm.get_connector_handler(fe.CONNECTOR_PERSIST)
        self.assertEqual(tbl.column_names, h.event_manager.get('task1_outcome').column_names)
        result = fe.load_persist_canonical()
        self.assertEqual(tbl.column_names, result.column_names)
        self.assertEqual(tbl.shape, result.shape)

    def test_run_intent_pipeline(self):
        fe = FeatureEngineer.from_env('test', has_contract=False)
        tools: FeatureEngineerIntent = fe.tools
        # reload the properties
        fe = FeatureEngineer.from_env('test')
        _ = tools.get_synthetic_data_types(size=10, extend=True)
        result = fe.tools.run_intent_pipeline()
        self.assertEqual((10, 19), result.shape)
        _ = tools.correlate_number(result, header='num')
        result = fe.tools.run_intent_pipeline(canonical=result)
        self.assertEqual((10, 20), result.shape)
        _ = tools.build_profiling(result, profiling='quality')
        pprint(pm_view('feature_build', 'test', 'intent'))
        result = fe.tools.run_intent_pipeline(canonical=result)
        self.assertEqual((21, 4), result.shape)

    def test_run_intent_pipeline_order(self):
        fe = FeatureEngineer.from_env('test', has_contract=False)
        tools: FeatureEngineerIntent = fe.tools
        # reload the properties
        fe = FeatureEngineer.from_env('test')
        _ = tools.get_synthetic_data_types(size=10, extend=True, intent_order=0)
        _ = tools.correlate_number(_, header='num', intent_order=1)
        _ = tools.build_profiling(_, profiling='quality', intent_order=1)
        pprint(pm_view('feature_build', 'test', 'intent'))
        result = fe.tools.run_intent_pipeline()
        self.assertEqual((21, 4), result.shape)

    def test_run_intent_pipeline_canonical(self):
        fe = FeatureEngineer.from_env('test', has_contract=False)
        tools: FeatureEngineerIntent = fe.tools
        # reload the properties
        fe = FeatureEngineer.from_env('test')
        tbl = tools.get_synthetic_data_types(size=10, extend=True, save_intent=False)
        _ = tools.correlate_number(tbl, header='num')
        _ = tools.build_profiling(_, profiling='quality')
        pprint(pm_view('feature_build', 'test', 'intent'))
        result = fe.tools.run_intent_pipeline(canonical=tbl)
        self.assertEqual((21, 4), result.shape)
        # get number of columns from the summary
        self.assertEqual(str(20), result.column('summary').slice(7, 1).to_pylist()[0])

    def test_run_intent_pipeline_intent_level(self):
        fe = FeatureEngineer.from_env('test', has_contract=False)
        tools: FeatureEngineerIntent = fe.tools
        # reload the properties
        fe = FeatureEngineer.from_env('test')
        tbl = tools.get_synthetic_data_types(size=10, extend=True, save_intent=False)
        _ = tools.correlate_number(tbl, header='num', intent_level='quantity')
        _ = tools.build_profiling(_, profiling='quality', intent_level='quantity')
        pprint(pm_view('feature_build', 'test', 'intent'))
        result = fe.tools.run_intent_pipeline(canonical=tbl, intent_level='quantity')
        self.assertEqual((21, 4), result.shape)
        # get number of columns from the summary
        self.assertEqual(str(20), result.column('summary').slice(7, 1).to_pylist()[0])

    #
    def test_get_noise(self):
        fe = FeatureEngineer.from_memory()
        tools: FeatureEngineerIntent = fe.tools
        tbl = tools.get_noise(10, num_columns=3)
        self.assertEqual((10, 3), tbl.shape)
        self.assertEqual(['A', 'B', 'C'], tbl.column_names)
        tbl = tools.get_noise(10, num_columns=3, name_prefix='P_')
        self.assertEqual(['P_A', 'P_B', 'P_C'], tbl.column_names)

    def test_correlate_replace(self):
        fe = FeatureEngineer.from_memory()
        tools: FeatureEngineerIntent = fe.tools
        tbl = tools.get_synthetic_data_types(1000)
        result = tools.correlate_replace(tbl, 'cat', 'INACTIVE', 'PAUSED', to_header='cat')
        print(result.column('cat').combine_chunks().dictionary)
        result = tools.correlate_replace(tbl, 'cat', 'IVE', 'XYZ', is_regex=True, to_header='cat')
        print(result.column('cat').combine_chunks().dictionary)

    def test_get_sample_map(self):
        fe = FeatureEngineer.from_memory()
        tools: FeatureEngineerIntent = fe.tools
        print(tools.sample_map)
        # tools.get_sample_map('us_persona', size=10, )
        i = tools.sample_inspect('us_persona')
        print(i)

    def test_raise(self):
        with self.assertRaises(KeyError) as context:
            env = os.environ['NoEnvValueTest']
        self.assertTrue("'NoEnvValueTest'" in str(context.exception))

def pm_view(capability: str, task: str, section: str=None):
    uri = os.path.join(os.environ['HADRON_PM_PATH'], f"hadron_pm_{capability}_{task}.parquet")
    tbl = pq.read_table(uri)
    tbl = tbl.column(0).combine_chunks()
    result = ast.literal_eval(tbl.to_pylist()[0]).get(capability,{}).get(task,{})
    return result.get(section, {}) if isinstance(section, str) and section in result.keys() else result


if __name__ == '__main__':
    unittest.main()



from ds_capability.components.commons import Commons
from ds_capability.intent.abstract_feature_intent import AbstractFeatureIntentModel
from ds_capability.managers.feature_engineer_property_manager import FeatureEngineerPropertyManager


class AbstractFeatureEngineerIntentModel(AbstractFeatureIntentModel):

    def __init__(self, property_manager: FeatureEngineerPropertyManager, default_save_intent: bool=None,
                 default_intent_level: [str, int, float]=None, order_next_available: bool=None,
                 default_replace_intent: bool=None):
        """initialisation of the Intent class.

        :param property_manager: the property manager class that references the intent contract.
        :param default_save_intent: (optional) The default action for saving intent in the property manager
        :param default_intent_level: (optional) the default level intent should be saved at
        :param order_next_available: (optional) if the default behaviour for the order should be next available order
        :param default_replace_intent: (optional) the default replace existing intent behaviour
        """
        default_save_intent = default_save_intent if isinstance(default_save_intent, bool) else True
        default_replace_intent = default_replace_intent if isinstance(default_replace_intent, bool) else True
        default_intent_level = default_intent_level if isinstance(default_intent_level, (str, int, float)) else 'primary'
        default_intent_order = -1 if isinstance(order_next_available, bool) and order_next_available else 0
        intent_param_exclude = ['canonical']
        intent_type_additions = []
        super().__init__(property_manager=property_manager, default_save_intent=default_save_intent,
                         intent_param_exclude=intent_param_exclude, default_intent_level=default_intent_level,
                         default_intent_order=default_intent_order, default_replace_intent=default_replace_intent,
                         intent_type_additions=intent_type_additions)
        self.label_gen = Commons.label_gen()



import unittest
import os
from pathlib import Path
import shutil
import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc
from ds_capability import FeatureEngineer
from ds_capability.intent.feature_engineer_intent import FeatureEngineerIntent
from ds_core.properties.property_manager import PropertyManager

# Pandas setup
pd.set_option('max_colwidth', 320)
pd.set_option('display.max_rows', 100)
pd.set_option('display.max_columns', 99)
pd.set_option('expand_frame_repr', True)


class SyntheticTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        # clean out any old environments
        for key in os.environ.keys():
            if key.startswith('HADRON'):
                del os.environ[key]
        # Local Domain Contract
        os.environ['HADRON_PM_PATH'] = os.path.join('working', 'contracts')
        os.environ['HADRON_PM_TYPE'] = 'json'
        # Local Connectivity
        os.environ['HADRON_DEFAULT_PATH'] = Path('working/data').as_posix()
        # Specialist Component
        try:
            os.makedirs(os.environ['HADRON_PM_PATH'])
        except OSError:
            pass
        try:
            os.makedirs(os.environ['HADRON_DEFAULT_PATH'])
        except OSError:
            pass
        try:
            shutil.copytree('../_test_data', os.path.join(os.environ['PWD'], 'working/source'))
        except OSError:
            pass
        PropertyManager._remove_all()

    def tearDown(self):
        try:
            shutil.rmtree('working')
        except OSError:
            pass

    def test_for_smoke(self):
        fe = FeatureEngineer.from_memory()
        tools: FeatureEngineerIntent = fe.tools
        tbl = tools.get_synthetic_data_types(10)
        old_schema = tbl.schema
        fe.add_connector_uri('sample', './working/data/sample.parquet')
        fe.save_canonical('sample', tbl)
        result = fe.load_canonical('sample')
        self.assertTrue(old_schema == result.schema)

    def test_from_env(self):
        fe = FeatureEngineer.from_env('tester', has_contract=False)
        tools: FeatureEngineerIntent = fe.tools
        fe.set_persist()
        tbl = tools.get_synthetic_data_types(10)
        old_schema = tbl.schema
        fe.add_connector_uri('sample', './working/data/sample.parquet')
        fe.save_canonical('sample', tbl)
        result = fe.load_canonical('sample')
        self.assertTrue(old_schema == result.schema)

    def test_from_source(self):
        fe = FeatureEngineer.from_env('tester', has_contract=False)
        tools: FeatureEngineerIntent = fe.tools
        fe.set_source('source/hadron_synth_origin.pq')
        fe.set_persist()
        fe.run_component_pipeline()
        result = fe.load_persist_canonical()
        print(result.column_names)



    def test_raise(self):
        with self.assertRaises(KeyError) as context:
            env = os.environ['NoEnvValueTest']
        self.assertTrue("'NoEnvValueTest'" in str(context.exception))


if __name__ == '__main__':
    unittest.main()



data = ['Annex', 'Arcade', 'Avenue', 'Bayou', 'Beach', 'Bend', 'Bluff', 'Bluffs', 'Bottom', 'Boulevard', 'Branch',
        'Bridge', 'Brook', 'Brooks', 'Burg', 'Burgs', 'Bypass', 'Camp', 'Canyon', 'Cape', 'Causeway', 'Center',
        'Centers', 'Circle', 'Circle', 'Cliff', 'Cliffs', 'Club', 'Common', 'Commons', 'Corner', 'Corners', 'Course',
        'Court', 'Courts', 'Cove', 'Coves', 'Creek', 'Crescent', 'Crest', 'Crossing', 'Crossroad', 'Curve', 'Dale',
        'Dam', 'Divide', 'Drive', 'Drives', 'Estate', 'Estates', 'Expressway', 'Extension', 'Extensions', 'Fall',
        'Falls', 'Ferry', 'Field', 'Fields', 'Flat', 'Flats', 'Ford', 'Fords', 'Forest', 'Forge', 'Forges', 'Fork',
        'Forks', 'Fort', 'Freeway', 'Garden', 'Gardens', 'Gateway', 'Glen', 'Glens', 'Green', 'Greens', 'Grove',
        'Groves', 'Harbor', 'Harbors', 'Haven', 'Heights', 'Highway', 'Hill', 'Hills', 'Hollow', 'Inlet', 'Island',
        'Islands', 'Isle', 'Junction', 'Junctions', 'Key', 'Keys', 'Knoll', 'Knolls', 'Lake', 'Lakes', 'Land',
        'Landing', 'Lane', 'Light', 'Lights', 'Loaf', 'Lock', 'Locks', 'Lodge', 'Loop', 'Mall', 'Manor', 'Manors',
        'Meadow', 'Meadows', 'Mews', 'Mill', 'Mills', 'Mission', 'Motorway', 'Mount', 'Mountain', 'Mountains', 'Neck',
        'Orchard', 'Oval', 'Overpass', 'Park', 'Parks', 'Parkway', 'Parkways', 'Pass', 'Passage', 'Path', 'Pike',
        'Pine', 'Pines', 'Place', 'Plain', 'Plains', 'Plaza', 'Point', 'Points', 'Port', 'Ports', 'Prairie', 'Radial',
        'Ramp', 'Ranch', 'Rapid', 'Rapids', 'Rest', 'Ridge', 'Ridges', 'River', 'Road', 'Roads', 'Route', 'Row', 'Rue',
        'Run', 'Shoal', 'Shoals', 'Shore', 'Shores', 'Skyway', 'Spring', 'Springs', 'Spur', 'Square', 'Squares',
        'Station', 'Stravenue', 'Stream', 'Street', 'Streets', 'Summit', 'Terrace', 'Throughway', 'Trace', 'Track',
        'Trafficway', 'Trail', 'Trailer', 'Tunnel', 'Turnpike', 'Underpass', 'Union', 'Unions', 'Valley', 'Valley',
        'Viaduct', 'View', 'Views', 'Village', 'Villages', 'Ville', 'Vista', 'Walk', 'Wall', 'Way', 'Well', 'Wells']



import unittest
import os
from pathlib import Path
import shutil
import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc
from ds_capability import FeatureBuild
from ds_capability.intent.feature_build_intent import FeatureBuildIntent
from ds_core.properties.property_manager import PropertyManager

# Pandas setup
pd.set_option('max_colwidth', 320)
pd.set_option('display.max_rows', 100)
pd.set_option('display.max_columns', 99)
pd.set_option('expand_frame_repr', True)


class FeatueBuildDiffTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        # clean out any old environments
        for key in os.environ.keys():
            if key.startswith('HADRON'):
                del os.environ[key]
        # Local Domain Contract
        os.environ['HADRON_PM_PATH'] = os.path.join('working', 'contracts')
        # os.environ['HADRON_PM_TYPE'] = 'json'
        # Local Connectivity
        os.environ['HADRON_DEFAULT_PATH'] = Path('working/data').as_posix()
        # Specialist Component
        try:
            os.makedirs(os.environ['HADRON_PM_PATH'])
        except OSError:
            pass
        try:
            os.makedirs(os.environ['HADRON_DEFAULT_PATH'])
        except OSError:
            pass
        try:
            shutil.copytree('../_test_data', os.path.join(os.environ['PWD'], 'working/source'))
        except OSError:
            pass
        PropertyManager._remove_all()

    def tearDown(self):
        try:
            shutil.rmtree('working')
        except OSError:
            pass

    def test_for_smoke(self):
        fb = FeatureBuild.from_memory()
        tools: FeatureBuildIntent = fb.tools
        fb.set_source_uri('./working/source/hadron_synth_origin.pq')
        fb.add_connector_uri('sample', './working/source/hadron_synth_other.pq')
        tbl = fb.load_source_canonical()
        result = tools.build_difference(tbl, other='sample', on_key='unique')
        for c in result.column_names:
            if c == 'num':
                self.assertEqual(5, pc.sum(result.column(c)).as_py())
            elif c == 'int':
                self.assertEqual(2, pc.sum(result.column(c)).as_py())
            elif c == 'unique':
                self.assertLess(result.num_rows, pc.sum(result.column(c)).as_py())
            else:
                self.assertEqual(0, pc.sum(result.column(c)).as_py())

    def test_pipeline(self):
        fb = FeatureBuild.from_env('tester', has_contract=False)
        tools: FeatureBuildIntent = fb.tools
        fb.set_source_uri('./working/source/hadron_synth_origin.pq')
        fb.add_connector_uri('sample', './working/source/hadron_synth_other.pq')
        tbl = fb.load_source_canonical()
        _ = tools.build_difference(tbl, other='sample', on_key='unique', intent_level='diff')
        fb.run_component_pipeline()
        result = fb.load_persist_canonical()
        self.assertEqual((10,7), result.shape)

    def test_model_no_difference(self):
        fb = FeatureBuild.from_memory()
        tools: FeatureBuildIntent = fb.tools
        fb.add_connector_uri('sample', './working/data/sample.parquet')
        tbl = tools.get_synthetic_data_types(10, inc_nulls=False)
        fb.save_canonical('sample', tbl)
        tbl = tools.build_difference(tbl, other='sample', on_key='int', drop_zero_sum=True)
        self.assertEqual(['int'], tbl.column_names)
        self.assertEqual(0, tbl.num_rows)

    def test_model_difference_num(self):
        fb = FeatureBuild.from_memory()
        tools: FeatureBuildIntent = fb.tools
        df = pa.table(data=    {"A": list("ABCDEFG"), "B": [1, 2, 5, 5, 3, 3, 1], 'C': [0,  2, 3, 3, 3, 2, 1], 'D': [0, 2, 0, 4, 3, 2, 1]})
        target = pa.table(data={"A": list("ABCDEFG"), "B": [1, 2, 5, 4, 3, 3, 1], 'C': [11, 2, 3, 4, 3, 2, 1], 'D': [0, 2, 0, 4, 3, 2, 1]})
        fb.add_connector_persist('target', uri_file='working/data/target.csv')
        fb.save_canonical('target', target)
        # normal
        result = tools.build_difference(df, 'target', on_key='A', drop_zero_sum=False)
        self.assertEqual((7,4), result.shape)
        # drop zero rows
        result = tools.build_difference(df, 'target', on_key='A', drop_zero_sum=True)
        self.assertEqual((2,3), result.shape)
        self.assertEqual(['A', 'D'], result.column('A').to_pylist())
        self.assertEqual([0,1], result.column('B').to_pylist())
        self.assertEqual([1,1], result.column('C').to_pylist())

    def test_model_difference_str(self):
        fb = FeatureBuild.from_memory()
        tools: FeatureBuildIntent = fb.tools
        df = pa.table(data=    {"A": list("ABCDEFG"), "B": ['B', 'C', 'A', 'A', 'F', 'E', 'G'], 'C': ['L', 'L',  'M', 'N', 'J', 'K', 'M']})
        target = pa.table(data={"A": list("ABCDEFG"), "B": ['B', 'C', 'D', 'A', 'F', 'E', 'G'], 'C': ['L', 'FX', 'M', 'N', 'P', 'K', 'M']})
        fb.add_connector_persist('target', uri_file='working/data/target.csv')
        fb.save_canonical('target', target)
        # tests
        result = tools.build_difference(df, 'target', on_key='A', drop_zero_sum=False)
        self.assertEqual((7,3), result.shape)
        result = tools.build_difference(df, 'target', on_key='A', drop_zero_sum=True)
        self.assertEqual((3,3), result.shape)
        self.assertEqual(['B', 'C', 'E'], result.column('A').to_pylist())
        self.assertEqual([0,1,0], result.column('B').to_pylist())
        self.assertEqual([1,0,1], result.column('C').to_pylist())

    def test_model_difference_unmatched(self):
        fb = FeatureBuild.from_memory()
        tools: FeatureBuildIntent = fb.tools
        df = pa.table(data={"X": list("ABCDEFGHK"),"Y": list("ABCABCABC"),  "B": ['B','C','A','A','F','E','G','X','Y'],'C': ['L','L','M','N','J','K','M','X','Y']})
        target = pa.table(data={"X": list("ABCDEFGX"),"Y": list("ABCABCAB"),"B": ['B','C','D','A','F','E','G','P'],    'C': ['L','FX','M','N','P','K','M','P'],"D": list("XYZXYZXY")})
        fb.add_connector_persist('target', uri_file='working/data/target.csv')
        fb.add_connector_uri('unmatched', uri='working/data/unmatched.csv')
        fb.save_canonical('target', target)
        # test
        self.assertEqual((9, 4), df.shape)
        self.assertEqual((8, 5), target.shape)
        result = tools.build_difference(df, 'target', on_key=['X', 'Y'], drop_zero_sum=True, unmatched_connector='unmatched')
        self.assertEqual((3,4), result.shape)
        unmatched = fb.load_canonical('unmatched')
        self.assertEqual((3,4), result.shape)
        self.assertEqual(['left_only', 'left_only', 'right_only', ], unmatched.column('found_in').to_pylist())
        self.assertEqual(['H', 'K', 'X', ], unmatched.column('X').to_pylist())

    def test_model_difference_unmatched_data(self):
        fb = FeatureBuild.from_memory()
        tools: FeatureBuildIntent = fb.tools
        fb.set_source_uri('working/source/hadron_synth_origin.pq')
        fb.add_connector_uri('target', uri='working/source/hadron_synth_other.pq')
        fb.add_connector_uri('unmatched', uri='working/data/unmatched.csv')
        # data
        df = fb.load_source_canonical()
        target = fb.load_canonical('target')
        self.assertEqual(((12, 10), (15, 8)), (df.shape, target.shape))
        # test
        _ = tools.build_difference(df, 'target', on_key=['unique'], drop_zero_sum=True, unmatched_connector='unmatched')
        unmatched = fb.load_canonical('unmatched')
        self.assertEqual((7, 12), unmatched.shape)
        column = unmatched.column('found_in')
        mapping = pc.equal(column, "left_only")
        self.assertEqual(2, column.filter(mapping).length())
        mapping = pc.equal(column, "right_only")
        self.assertEqual(5, column.filter(mapping).length())

    def test_model_difference_equal(self):
        fb = FeatureBuild.from_memory()
        tools: FeatureBuildIntent = fb.tools
        df = pa.table(data=    {"X": list("ABCDEFG"), "Y": list("RSTUVWX"), "B": ['B', 'C', 'A', 'A', 'F', 'E', 'G'], 'C': ['L', 'L',  'M', 'N', 'J', 'K', 'M']})
        target = pa.table(data={"X": list("ABCDEFG"), "Y": list("RSTUVWX"), "B": ['B', 'C', 'A', 'A', 'F', 'E', 'G'], 'C': ['L', 'L',  'M', 'N', 'J', 'K', 'M']})
        fb.add_connector_persist('target', uri_file='working/data/target.csv')
        fb.save_canonical('target', target)
        # identical
        result = tools.build_difference(df, 'target', on_key='X')
        self.assertEqual((7,4), result.shape)
        result = tools.build_difference(df, 'target', on_key='X', drop_zero_sum=True)
        self.assertEqual((0,1), result.shape)
        self.assertEqual(['X'], result.column_names)
        result = tools.build_difference(df, 'target', on_key=['X', 'Y'], drop_zero_sum=True)
        self.assertEqual((0,2), result.shape)
        self.assertCountEqual(['Y','X'], result.column_names)

    def test_model_difference_multi_key(self):
        fb = FeatureBuild.from_memory()
        tools: FeatureBuildIntent = fb.tools
        df = pa.table(data=    {"X": list("ABCDEFG"), "Y": list("RSTUVWX"), "B": ['B', 'C', 'A', 'A', 'F', 'E', 'G'], 'C': ['L', 'L',  'M', 'N', 'J', 'K', 'M']})
        target = pa.table(data={"X": list("ABCDEFG"), "Y": list("RSTUVWX"), "B": ['B', 'C', 'D', 'A', 'F', 'E', 'G'], 'C': ['L', 'FX', 'M', 'N', 'P', 'K', 'M']})
        fb.add_connector_persist('target', uri_file='working/data/target.csv')
        fb.save_canonical('target', target)
        # one key
        result = tools.build_difference(df, 'target', on_key=['X'])
        self.assertTrue(all((v is None) or isinstance(v, str) for v in result.column('X').to_pylist()))
        self.assertFalse(all((v is None) or isinstance(v, str) for v in result.column('Y').to_pylist()))
        # two keys
        result = tools.build_difference(df, 'target', on_key=['X', 'Y'])
        self.assertTrue(all((v is None) or isinstance(v, str) for v in result.column('X').to_pylist()))
        self.assertTrue(all((v is None) or isinstance(v, str) for v in result.column('Y').to_pylist()))


    def test_model_difference_order(self):
        fb = FeatureBuild.from_memory()
        tools: FeatureBuildIntent = fb.tools
        df = pa.table(data={"A": list("ABCDEFG"), "B": list("ABCFCBA"), 'C': list("BCDECFB"), 'D': [0, 2, 0, 4, 3, 2, 1]})
        target = pa.table(data={"A": list("ABCDEFG"), "B": list("BBCDCAA"), 'C': list("BCDECFB"), 'D': [0, 2, 0, 4, 1, 2, 1]})
        _ = target.to_pandas().sample(frac = 1)
        target = pa.Table.from_pandas(_)
        fb.add_connector_persist('target', uri_file='working/data/target.csv')
        fb.save_canonical('target', target)
        result = tools.build_difference(df, 'target', on_key='A', drop_zero_sum=True)
        self.assertEqual((4,3), result.shape)
        self.assertEqual(['A', 'B', 'D'], result.column_names)

    def test_model_difference_drop(self):
        fb = FeatureBuild.from_memory()
        tools: FeatureBuildIntent = fb.tools
        df = pa.table(data={"A": list("ABCDEFG"), "B": list("ABCFCBA"), 'C': list("BCDECFB"), 'D': [0, 2, 0, 4, 3, 2, 1]})
        target = pa.table(data={"A": list("ABCDEFG"), "B": list("BBCDCAA"), 'C': list("BCDECFB"), 'D': [0, 2, 0, 4, 1, 2, 1]})
        fb.add_connector_persist('target', uri_file='working/data/target.csv')
        fb.save_canonical('target', target)
        result = tools.build_difference(df, 'target', on_key='A', drop_zero_sum=True)
        self.assertEqual((4,3), result.shape)
        self.assertEqual(['A', 'B', 'D'], result.column_names)

    def test_model_difference_summary(self):
        fb = FeatureBuild.from_memory()
        tools: FeatureBuildIntent = fb.tools
        df = pa.table(data={"X":  list("ABCDEFG"),    "Y": list("ABCDEFG"), "B": [1, 2, 3, 4, 3, 3, 1], 'C': [0, 2, 0, 4, 3, 2, 1]})
        target = pa.table(data={"X": list("ABCDEFG"), "Y": list("ABCDEFG"), "B": [1, 2, 5, 4, 3, 3, 1], 'C': [1, 2, 3, 4, 3, 2, 1]})
        fb.add_connector_persist('target', uri_file='target.csv')
        fb.save_canonical('target', target)
        # summary connector
        fb.add_connector_persist('summary', uri_file='summary.csv')
        _ = tools.build_difference(df, 'target', on_key='X', summary_connector='summary')
        result = fb.load_canonical('summary')
        self.assertEqual(result.shape, (6,2))
        self.assertEqual(result.column('Attribute').to_pylist(), ['matching','left_only','right_only','B','C','Y'])
        self.assertEqual(result.column('Summary').to_pylist(), [7,0,0,1,2,0])
        _ = tools.build_difference(df, 'target', on_key='X', summary_connector='summary', drop_zero_sum=True)
        result = fb.load_canonical('summary')
        self.assertEqual(result.shape, (5,2))
        self.assertEqual(result.column('Attribute').to_pylist(), ['matching','left_only','right_only','B','C'])
        self.assertEqual(result.column('Summary').to_pylist(), [7,0,0,1,2])

    def test_model_difference_detail(self):
        fb = FeatureBuild.from_memory()
        tools: FeatureBuildIntent = fb.tools
        df = pa.table(data={"X":  list("CBADEFG"), "Y":  list("ABCDEFG"), "B": [1, 2, 3, 4, 3, 3, 1], 'C': [0, 2, 0, 4, 3, 2, 1]})
        target = pa.table(data={"X": list("CBADEFG"), "Y":  list("ABCDEFG"), "B": [1, 2, 5, 4, 3, 3, 1], 'C': [1, 2, 3, 4, 3, 2, 1]})
        fb.add_connector_persist('target', uri_file='working/data/target.csv')
        fb.save_canonical('target', target)
        # detail connector
        fb.add_connector_persist('detail', uri_file='detail.csv')
        _ = tools.build_difference(df, 'target', on_key='X', detail_connector='detail')
        result = fb.load_canonical('detail')
        self.assertEqual(result.shape, (2,5))
        self.assertEqual(result.column_names, ['X', 'B_x', 'B_y', 'C_x', 'C_y'])
        # self.assertEqual(result.loc[0].values.tolist(), ['A', '3', '5', 0, 3])
        _ = tools.build_difference(df, 'target', on_key=['X', 'Y'], detail_connector='detail')
        result = fb.load_canonical('detail')
        self.assertEqual(result.shape, (2,6))
        self.assertEqual(result.column_names, ['X', 'Y', 'B_x', 'B_y', 'C_x', 'C_y'])
        # self.assertEqual(result.loc[0].values.tolist(), ['A', 'C', '3', '5', 0, 3])


    def test_raise(self):
        with self.assertRaises(KeyError) as context:
            env = os.environ['NoEnvValueTest']
        self.assertTrue("'NoEnvValueTest'" in str(context.exception))


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from ds_capability.components.commons import Commons
from ds_capability.intent.feature_build_intent import FeatureBuildIntent
from ds_capability.managers.feature_build_property_manager import FeatureBuildPropertyManager
from ds_capability.components.abstract_common_component import AbstractCommonComponent

__author__ = 'Darryl Oatridge'


class FeatureBuild(AbstractCommonComponent):

    @classmethod
    def from_uri(cls, task_name: str, uri_pm_path: str, creator: str, uri_pm_repo: str=None, pm_file_type: str=None,
                 pm_module: str=None, pm_handler: str=None, pm_kwargs: dict=None, default_save=None,
                 reset_templates: bool=None, template_path: str=None, template_module: str=None,
                 template_source_handler: str=None, template_persist_handler: str=None, align_connectors: bool=None,
                 default_save_intent: bool=None, default_intent_level: bool=None, order_next_available: bool=None,
                 default_replace_intent: bool=None, has_contract: bool=None) -> FeatureBuild:
        """ Class Factory Method to instantiates the component's application. The Factory Method handles the
        instantiation of the Properties Manager, the Intent Model and the persistence of the uploaded properties.
        See class inline _docs for an example method

         :param task_name: The reference name that uniquely identifies a task or subset of the property manager
         :param uri_pm_path: A URI that identifies the resource path for the property manager.
         :param creator: A user name for this task activity.
         :param uri_pm_repo: (optional) A repository URI to initially load the property manager but not save to.
         :param pm_file_type: (optional) defines a specific file type for the property manager
         :param pm_module: (optional) the module or package name where the handler can be found
         :param pm_handler: (optional) the handler for retrieving the resource
         :param pm_kwargs: (optional) a dictionary of kwargs to pass to the property manager
         :param default_save: (optional) if the configuration should be persisted. default to 'True'
         :param reset_templates: (optional) reset connector templates from environ variables. Default True
                                (see `report_environ()`)
         :param template_path: (optional) a template path to use if the environment variable does not exist
         :param template_module: (optional) a template module to use if the environment variable does not exist
         :param template_source_handler: (optional) a template source handler to use if no environment variable
         :param template_persist_handler: (optional) a template persist handler to use if no environment variable
         :param align_connectors: (optional) resets aligned connectors to the template. default Default True
         :param default_save_intent: (optional) The default action for saving intent in the property manager
         :param default_intent_level: (optional) the default level intent should be saved at
         :param order_next_available: (optional) if the default behaviour for the order should be next available order
         :param default_replace_intent: (optional) the default replace existing intent behaviour
         :param has_contract: (optional) indicates the instance should have a property manager domain contract
         :return: the initialised class instance
         """
        pm_file_type = pm_file_type if isinstance(pm_file_type, str) else 'json'
        pm_module = pm_module if isinstance(pm_module, str) else cls.DEFAULT_MODULE
        pm_handler = pm_handler if isinstance(pm_handler, str) else cls.DEFAULT_PERSIST_HANDLER
        _pm = FeatureBuildPropertyManager(task_name=task_name, creator=creator)
        _intent_model = FeatureBuildIntent(property_manager=_pm, default_save_intent=default_save_intent,
                                           default_intent_level=default_intent_level,
                                           order_next_available=order_next_available,
                                           default_replace_intent=default_replace_intent)
        super()._init_properties(property_manager=_pm, uri_pm_path=uri_pm_path, default_save=default_save,
                                 uri_pm_repo=uri_pm_repo, pm_file_type=pm_file_type, pm_module=pm_module,
                                 pm_handler=pm_handler, pm_kwargs=pm_kwargs, has_contract=has_contract)
        return cls(property_manager=_pm, intent_model=_intent_model, default_save=default_save,
                   reset_templates=reset_templates, template_path=template_path, template_module=template_module,
                   template_source_handler=template_source_handler, template_persist_handler=template_persist_handler,
                   align_connectors=align_connectors)

    @property
    def pm(self) -> FeatureBuildPropertyManager:
        return self._component_pm

    @property
    def intent_model(self) -> FeatureBuildIntent:
        return self._intent_model

    @property
    def tools(self) -> FeatureBuildIntent:
        return self._intent_model




from ds_core.properties.abstract_properties import AbstractPropertyManager

__author__ = 'Darryl Oatridge'


class FeatureSelectPropertyManager(AbstractPropertyManager):

    def __init__(self, task_name: str, creator: str):
        # set additional keys
        root_keys = [{'provenance': ['title', 'domain', 'description', 'license', 'provider', 'author', 'cost']}]
        knowledge_keys = ['observations', 'actions', 'attributes']
        super().__init__(task_name=task_name, root_keys=root_keys, knowledge_keys=knowledge_keys, creator=creator)

    @property
    def provenance(self) -> dict:
        """Return the provenance report_canonical"""
        return self.get(self.KEY.provenance_key, {})

    def report_provenance(self) -> dict:
        """Return the provenance report_canonical"""
        report = dict()
        for catalog in self.get(self.KEY.provenance_key, {}).keys():
            _key = self.join(self.KEY.provenance_key, catalog)
            if catalog in ['provider', 'author', 'cost', 'license']:
                for name, value in self.get(_key, {}).items():
                    report[f"{catalog}_{name}"] = value
            else:
                report[catalog] = self.get(_key, '')
        return report

    def set_provenance(self, title: str=None, domain: str=None, description: str=None, license_type: str=None,
                       license_name: str=None, license_uri: str=None, cost_price: str=None, cost_code: str=None,
                       cost_type: str=None, provider_name: str=None, provider_uri: str=None, provider_note: str=None,
                       author_name: str=None, author_uri: str=None, author_contact: str=None):
        """ sets the provenance values. Only sets those passed

        :param title: (optional) the title of the provenance
        :param domain: (optional) the domain it sits within
        :param description: (optional) a description of the provenance
        :param license_type: (optional) The type of the license such as PDDL, ODC-By, ODC-ODbL
        :param license_name: (optional) The full name of the license e.g. 'Open Data Commons Attribution License'
        :param license_uri: (optional) The link to the license e.g. 'https://opendatacommons.org/licenses/by/summary/'
        :param cost_price: (optional) a cost price associated with this provenance
        :param cost_code: (optional) a cost centre code or reference code
        :param cost_type: (optional) the cost type or description
        :param provider_name: (optional) the provider system or institution name or title
        :param provider_uri: (optional) a uri reference that helps identify the provider
        :param provider_note: (optional) any notes that might be useful
        :param author_name: (optional) the author of the data
        :param author_uri: (optional) the author uri
        :param author_contact: (optional)the the author contact information
        """
        params = locals().copy()
        params.pop('self', None)
        for name, value in params.items():
            if value is None:
                continue
            if name in ['title', 'domain', 'description']:
                self.set(self.join(self.KEY.provenance_key, name), value)
            if name.startswith('provider') or name.startswith('author') or name.startswith('cost') or name.startswith(
                    'license'):
                key, item = name.split(sep='_')
                self.set(self.join(self.KEY.provenance_key, key, item), value)
        return

    def reset_provenance(self):
        """resets provenance back to its default values"""
        self._base_pm.remove(self.KEY.provenance_key)
        self.set(self.KEY.provenance_key, {})
        return



# Developing Mongo Persist Handler
import ast
import pyarrow as pa

from ds_core.handlers.abstract_handlers import AbstractSourceHandler, AbstractPersistHandler
from ds_core.handlers.abstract_handlers import HandlerFactory, ConnectorContract
from ds_capability.components.commons import Commons


class MongodbSourceHandler(AbstractSourceHandler):
    """ A mongoDB source handler"""

    def __init__(self, connector_contract: ConnectorContract):
        """ initialise the Handler passing the source_contract dictionary """
        # required module import
        self.mongo = HandlerFactory.get_module('pymongo')
        self.pymongo = HandlerFactory.get_module('pymongoarrow')
        self.bson = HandlerFactory.get_module('bson')
        super().__init__(connector_contract)

        _kwargs = {**self.connector_contract.query, **self.connector_contract.query}
        database = _kwargs.pop('database', "hadron_db")
        self.collection_name = _kwargs.pop('collection', "records")
        self._mongo_find = ast.literal_eval(_kwargs.pop('find').replace("'", '"')) if _kwargs.get('find') else {}
        self._mongo_aggregate = ast.literal_eval(_kwargs.pop('aggregate').replace("'", '"')) if _kwargs.get('aggregate') else None
        self._mongo_project = ast.literal_eval(_kwargs.pop('project').replace("'", '"')) if _kwargs.get('project') else None
        self._mongo_limit = int(_kwargs.pop('limit').replace("'", '"')) if _kwargs.get('limit') else None
        self._mongo_skip = int(_kwargs.pop('skip').replace("'", '"')) if _kwargs.get('skip') else None
        self._mongo_sort = ast.literal_eval(_kwargs.pop('sort').replace("'", '"')) if _kwargs.get('sort') else None
        self._decode = _kwargs.pop('decode') if _kwargs.get('decode') else False

        self._if_exists = _kwargs.pop('if_exists', 'replace')
        self._file_state = 0
        self._changed_flag = True

        self._mongo_document = self.mongo.MongoClient(self.connector_contract.address)[database]
        self._mongo_collection = self._mongo_document[self.collection_name]

    def supported_types(self) -> list:
        """ The source types supported with this module"""
        return ['mongodb']

    def load_canonical(self, **kwargs) -> pa.Table:
        """ returns the canonical dataset based on the source contract
            The canonical in this instance is a dictionary that has the headers as the key and then
            the ordered list of values for that header
        """
        if not isinstance(self.connector_contract, ConnectorContract):
            raise ValueError("The PandasSource Connector Contract has not been set")

        def traverse_it(it):
            if isinstance(it, list):
                for item in it:
                    traverse_it(item)
            elif isinstance(it, dict):
                if '_id' in it.keys():
                    if isinstance(it.get('_id'), self.bson.ObjectId):
                        it.update({'_id': str(it.get('_id'))})
                for key in it.keys():
                    traverse_it(it[key])

        #    else: # returns top of the tree item

        if self._mongo_aggregate is not None:
            result = list(self._mongo_collection.aggregate(self._mongo_aggregate))
            if isinstance(result[0].get('_id'), self.bson.ObjectId) or self._decode:
                traverse_it(result)
            return Commons.table_flatten(pa.Table.from_pylist(result))
        cursor = self._mongo_collection.find(self._mongo_find, self._mongo_project)
        if self._mongo_limit is not None:
            cursor.limit(self._mongo_limit)
        if self._mongo_skip is not None:
            cursor.skip(self._mongo_skip)
        if self._mongo_sort is not None:
            cursor.sort(self._mongo_sort)
        result = list(cursor)
        if isinstance(result[0].get('_id'), self.bson.ObjectId) or self._decode:
            traverse_it(result)
        return Commons.table_flatten(pa.Table.from_pylist(result))

    def exists(self) -> bool:
        """ returns True if the collection exists """
        return self.collection_name in self._mongo_document.list_collection_names()

    def has_changed(self) -> bool:
        """ returns the amount of documents in the collection
            ... if the counts change ... then the collection was probably modified ...
            ... this assumes that records are never edited/updated ... nor deleted ...
        """
        _cc = self.connector_contract
        state = self._mongo_collection.count_documents(self._mongo_find)
        if state != self._file_state:
            self._changed_flag = True
            self._file_state = state
        return self._changed_flag

    def reset_changed(self, changed: bool = False):
        """ manual reset to say the file has been seen. This is automatically called if the file is loaded"""
        changed = changed if isinstance(changed, bool) else False
        self._changed_flag = changed


class MongodbPersistHandler(MongodbSourceHandler, AbstractPersistHandler):
    # a mongoDB persist handler

    def persist_canonical(self, canonical: pa.Table, **kwargs) -> bool:
        """ persists the canonical dataset
        """
        if not isinstance(self.connector_contract, ConnectorContract):
            return False
        _uri = self.connector_contract.uri
        return self.backup_canonical(canonical=canonical, collection_name=self.collection_name, **kwargs)

    def backup_canonical(self, canonical: pa.Table, collection_name: str, **kwargs) -> bool:
        """  creates a backup of the canonical to an alternative table """
        if not isinstance(self.connector_contract, ConnectorContract):
            return False
        _collection = self._mongo_document[collection_name]
        tbl = Commons.table_nest(canonical)
        # remove anything there as this is replaced
        _collection.delete_many({})
        resp = _collection.insert_many(tbl)
        return resp.acknowledged

    def remove_canonical(self) -> bool:
        if not isinstance(self.connector_contract, ConnectorContract):
            return False
        return self._mongo_document.drop_collection(self.collection_name)



data = ['Office Assistant II', 'Graphic Designer', 'Database Administrator III', 'Marketing Manager',
        'Computer Systems Analyst IV', 'Recruiter', 'Librarian', 'Food Chemist', 'VP Marketing',
        'Community Outreach Specialist', 'Sales Associate', 'Human Resources Assistant IV', 'Senior Sales Associate',
        'Help Desk Operator', 'Speech Pathologist', 'Occupational Therapist', 'Technical Writer', 'Analyst Programmer',
        'Research Assistant III', 'Quality Control Specialist', 'Junior Executive', 'Actuary', 'Health Coach III',
        'Cost Accountant', 'Internal Auditor', 'Dental Hygienist', 'Design Engineer', 'Accounting Assistant I',
        'Web Developer II', 'Editor', 'Web Designer I', 'Accounting Assistant I', 'Automation Specialist I',
        'Account Coordinator', 'Programmer Analyst I', 'Help Desk Operator', 'Social Worker', 'Senior Editor',
        'Registered Nurse', 'Safety Technician III', 'Internal Auditor', 'Nurse Practicioner',
        'Human Resources Manager', 'Assistant Manager', 'Research Nurse', 'VP Sales', 'Programmer Analyst I',
        'Accountant I', 'Clinical Specialist', 'Human Resources Manager', 'Editor', 'Data Coordiator',
        'Computer Systems Analyst II', 'Tax Accountant', 'Accountant II', 'Social Worker', 'Senior Sales Associate',
        'Accounting Assistant IV', 'Director of Sales', 'Senior Quality Engineer', 'Payment Adjustment Coordinator',
        'Paralegal', 'Actuary', 'Environmental Tech', 'Structural Engineer', 'Design Engineer',
        'Quality Control Specialist', 'Speech Pathologist', 'Quality Control Specialist', 'Help Desk Technician',
        'Software Engineer IV', 'VP Accounting', 'Account Executive', 'Desktop Support Technician', 'Junior Executive',
        'Financial Advisor', 'Database Administrator III', 'Developer I', 'Food Chemist', 'Information Systems Manager',
        'Executive Secretary', 'Cost Accountant', 'Director of Sales', 'Financial Analyst', 'Librarian',
        'Statistician III', 'Web Developer III', 'Structural Analysis Engineer', 'Desktop Support Technician',
        'Community Outreach Specialist', 'Research Assistant II', 'Assistant Professor', 'Safety Technician IV',
        'VP Product Management', 'Junior Executive', 'Accountant II', 'Design Engineer', 'Nuclear Power Engineer',
        'Operator']



import unittest
import os
from datetime import datetime
from pathlib import Path
import shutil
import pyarrow as pa
import pyarrow.compute as pc
from ds_core.handlers.abstract_handlers import ConnectorContract
from ds_core.properties.property_manager import PropertyManager
from ds_capability import FeatureBuild


class MongodbHandlerTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        # clean out any old environments
        for key in os.environ.keys():
            if key.startswith('HADRON'):
                del os.environ[key]
        # Local Domain Contract
        os.environ['HADRON_PM_PATH'] = os.path.join('working', 'contracts')
        os.environ['HADRON_PM_TYPE'] = 'json'
        # Local Connectivity
        os.environ['HADRON_DEFAULT_PATH'] = Path('working/data').as_posix()
        # Specialist Component
        try:
            os.makedirs(os.environ['HADRON_PM_PATH'])
        except OSError:
            pass
        try:
            os.makedirs(os.environ['HADRON_DEFAULT_PATH'])
        except OSError:
            pass
        PropertyManager._remove_all()

    def tearDown(self):
        try:
            shutil.rmtree('working')
        except OSError:
            pass

    def test_handler_default(self):
        fb = FeatureBuild.from_memory()
        tbl = fb.tools.get_synthetic_data_types(size=1_000)
        tbl = fb.tools.get_number(1_000_000, 10_000_000, canonical=tbl, at_most=1, size=tbl.num_rows, column_name='_id')
        fb.set_persist_uri("mongodb://admin:admin@localhost:27017/")
        fb.remove_canonical(fb.CONNECTOR_PERSIST)
        self.assertFalse(fb.pm.get_connector_handler(fb.CONNECTOR_PERSIST).exists())
        fb.save_persist_canonical(tbl)
        result = fb.load_persist_canonical()
        self.assertTrue(fb.pm.get_connector_handler(fb.CONNECTOR_PERSIST).exists())
        self.assertEqual((1000, 7), result.shape)
        self.assertEqual(['_id', 'cat', 'num', 'int', 'bool', 'date', 'string'], result.column_names)
        fb.remove_canonical(fb.CONNECTOR_PERSIST)

    def test_handler_sort(self):
        fb = FeatureBuild.from_memory()
        tbl = fb.tools.get_synthetic_data_types(size=10, seed=101)
        uri = "mongodb://admin:admin@localhost:27017/?project={'cat':1, 'num':1, '_id':0}&sort=[('num', 1)]"
        fb.set_persist_uri(uri=uri).save_persist_canonical(tbl)
        result = fb.load_persist_canonical()
        c = result.column('num').combine_chunks()
        self.assertEqual(pc.min(c), c[0])
        self.assertEqual(pc.max(c), c.slice(len(c)-1)[0])

    def test_handler_query(self):
        fb = FeatureBuild.from_memory()
        tbl = fb.tools.get_synthetic_data_types(size=1_000)
        os.environ['collection'] = 'hadron_table'
        uri = "mongodb://admin:admin@localhost:27017/?collection=${collection}&find={}&project={'cat':1, 'num':1}"
        fb.set_persist_uri(uri=uri)
        fb.remove_canonical(fb.CONNECTOR_PERSIST)
        fb.save_persist_canonical(tbl)
        result = fb.load_persist_canonical()
        self.assertEqual((1000, 2), result.shape)
        self.assertEqual(['cat', 'num'], result.column_names)
        fb.remove_canonical(fb.CONNECTOR_PERSIST)

    def test_handler_find_limit(self):
        fb = FeatureBuild.from_memory()
        tbl = fb.tools.get_synthetic_data_types(size=1_000)
        os.environ['collection'] = 'hadron_table'
        uri = "mongodb://admin:admin@localhost:27017/?collection=${collection}&find={}&project={'cat':1, 'num':1, '_id':0}&limit=2&spip=2"
        fb.set_persist_uri(uri=uri)
        fb.remove_canonical(fb.CONNECTOR_PERSIST)
        fb.save_persist_canonical(tbl)
        result = fb.load_persist_canonical()
        self.assertEqual((2, 2), result.shape)
        self.assertEqual(['cat', 'num'], result.column_names)
        fb.remove_canonical(fb.CONNECTOR_PERSIST)

    def test_handler_aggregate(self):
        os.environ['HADRON_AGG'] = """
        [
            {"$match": { 
                "cat": { "$eq": "ACTIVE" },
            }},
            {"$project": {
                "_id":0,
                "cat":1,  
                "num":1, 
                "int":1, 
            }},
            {'$limit':50}
        ]
        """
        fb = FeatureBuild.from_memory()
        tbl = fb.tools.get_synthetic_data_types(size=1_000)
        uri = "mongodb://admin:admin@localhost:27017/?aggregate=${HADRON_AGG}"
        fb.set_persist_uri(uri=uri)
        fb.remove_canonical(fb.CONNECTOR_PERSIST)
        fb.save_persist_canonical(tbl)
        result = fb.load_persist_canonical()
        self.assertEqual(['cat', 'num', 'int'], result.column_names)
        self.assertEqual(['ACTIVE'], pc.unique(result.column('cat')).to_pylist())
        self.assertEqual(50, result.num_rows)

    def test_connector_contract(self):
        os.environ['HADRON_ADDITION'] = 'myAddition'
        os.environ['HADRON_AGG'] = """
        [
            {"$match": { 
                "creditProfile": { "$ne": None },
                #"creditProfile.creditScore": { "$ne": None }
            }},
            {"$project": {
                "customerCategory":1,  
                "customerSubCategory":1, 
                "customerType":1, 
                "organizationChildRelationship": {
                    "$cond": {
                        "if": {"$isArray": "$organizationChildRelationship"},
                        "then": "$organizationChildRelationship",
                        "else": [
                            {
                                "organization": {
                                    "_id": None,
                                    "referredType": None
                                },
                                "relationshipType": None
                            }
                        ]
                    }},
                "riskCategory":1, 
                "status": 1,
                "creditProfile": 1, 
                "deniedList.isDenied": 1, 
                "deniedList.validFor.startDateTime": 1,
                "engagedParty": 1
            }},
            {'$limit':2000}
        ]
        """
        uri = "mongodb://admin:admin@localhost:27017/path1/path2/file.parquet?database=hadron_docs&collection=records&aggregate=${HADRON_AGG}"
        cc = ConnectorContract(uri=uri, module_name='', handler='', addition='${HADRON_ADDITION}')
        print(f"raw_uri = {cc.raw_uri}")
        print(f"uri = {cc.uri}")
        print(f"raw_kwargs = {cc.raw_kwargs}")
        print(f"address = {cc.address}")
        print(f"schema = {cc.schema}")
        print(f"hostname = {cc.hostname}")
        print(f"port = {cc.port}")
        print(f"username = {cc.username}")
        print(f"password = {cc.password}")
        print(f"path = {cc.path}")
        print(f"database = {cc.path[1:]}")
        print(f"query")
        extra = cc.query.pop('extra', None)
        print(f" extra = {extra}")
        find = cc.query.pop('find', None)
        print(f" mongo_find = {find}")
        aggregate = cc.query.pop('aggregate', None)
        print(f" mongo_aggregate = {aggregate}")
        collection = cc.query.pop('collection', None)
        print(f" collection = {collection}")
        print(f"kwargs")
        addition = cc.kwargs.get('addition', None)
        print(f" addition = {addition}")

    def test_raise(self):
        startTime = datetime.now()
        with self.assertRaises(KeyError) as context:
            env = os.environ['NoEnvValueTest']
        self.assertTrue("'NoEnvValueTest'" in str(context.exception))
        print(f"Duration - {str(datetime.now() - startTime)}")



if __name__ == '__main__':
    unittest.main()



import pandas as pd
import numpy as np
from sklearn.preprocessing import StandardScaler, LabelEncoder, OneHotEncoder
from sklearn.impute import SimpleImputer, KNNImputer
from sklearn.linear_model import LinearRegression
from sklearn.pipeline import Pipeline


class AutoPreprocessor:
    def __init__(self):
        self.datetime_columns = []  # Initialize datetime_columns as an empty list

    # Function to clean column names
    def clean_column_names(self, data):
        data.columns = data.columns.str.replace(' ', '_')  # Replace spaces with underscores
        data.columns = data.columns.str.strip()
        # Convert to lowercase
        data.columns = data.columns.str.lower()  # Remove leading and trailing whitespace, including newline characters
        data.columns = data.columns.str.replace(r'[^\w\s]', '', regex=True)  # Remove special characters
        return data

    def handle_missing_values(self, data, numerical_impute_method, categorical_impute_method):
        # Check for and drop columns exceeding the missing value threshold
        threshold = 0.7  # Set the threshold for missing values (70% or more)
        threshold_count = len(data) * threshold
        columns_to_drop = [col for col in data.columns if data[col].isna().sum() >= threshold_count]
        data = data.drop(columns=columns_to_drop)

        if not data.empty:
            numerical_data = data.select_dtypes(include=np.number)
            categorical_data = data.select_dtypes(include='object')

            if not numerical_data.empty:
                data = self._impute_numerical(data, numerical_impute_method)

            if not categorical_data.empty:
                data = self._impute_categorical(data, categorical_impute_method)

        return data

    def _impute_categorical(self, data, categorical_impute_method):
        categorical_columns = data.select_dtypes(include='object').columns

        for col in categorical_columns:
            if data[col].isna().sum() > 0:
                if col not in self.datetime_columns:  # Skip datetime columns
                    if categorical_impute_method == 'mode':
                        mode = data[col].mode().iloc[0]
                        data[col].fillna(mode, inplace=True)
                    elif categorical_impute_method == 'knn':
                        # Check if the column is numeric (can be converted to int)
                        if pd.api.types.is_numeric_dtype(data[col]):
                            imputer = KNNImputer(n_neighbors=5)
                            data[col] = imputer.fit_transform(data[col].values.reshape(-1, 1))
                            data[col] = data[col].astype(int)  # Convert back to integer after KNN imputation

        return data

    def _impute_numerical(self, data, numerical_impute_method):
        numerical_columns = data.select_dtypes(include=np.number).columns

        for col in numerical_columns:
            if data[col].isna().sum() > 0:
                if col not in self.datetime_columns:  # Skip datetime columns
                    if numerical_impute_method == 'auto':
                        # Choose imputation method automatically based on data characteristics
                        if len(data[col].unique()) > 10:
                            # Use K-NN for columns with many unique values
                            imputer = KNNImputer(n_neighbors=5)
                        else:
                            # Use Linear Regression for columns with fewer unique values
                            imputer = Pipeline([
                                ('scaler', StandardScaler()),
                                ('regressor', LinearRegression())
                            ])

                    elif numerical_impute_method == 'knn':
                        imputer = KNNImputer(n_neighbors=5)

                    elif numerical_impute_method in ['mean', 'median']:
                        imputer = SimpleImputer(strategy=numerical_impute_method)

                    data[col] = imputer.fit_transform(data[col].values.reshape(-1, 1))

        return data

    def preprocess(self, data, datetime_columns=None, numerical_impute_method='auto', categorical_impute_method='mode',
                   categorical_encoding_method='one-hot'):
        data = self.clean_column_names(data)  # Clean column names first

        # Set the datetime_columns attribute
        if datetime_columns:
            self.datetime_columns = datetime_columns

        data = self.handle_missing_values(data, numerical_impute_method, categorical_impute_method)

        if categorical_encoding_method == 'label':
            data = self.encode_categorical_label(data)
        elif categorical_encoding_method == 'one-hot':
            data = self.encode_categorical_one_hot(data)
        elif categorical_encoding_method == 'pd-dummies':
            data = self.encode_categorical_pd_dummies(data)

        data = self.scale_numerical(data)

        if datetime_columns:
            data = self.convert_datetime_columns(data, datetime_columns)

        return data

    def encode_categorical_label(self, data):
        # Encode categorical columns using LabelEncoder
        categorical_columns = data.select_dtypes(include='object').columns
        label_encoder = LabelEncoder()
        for col in categorical_columns:
            data[col] = label_encoder.fit_transform(data[col])
        return data

    def encode_categorical_one_hot(self, data):
        # Encode categorical columns using OneHotEncoder
        categorical_columns = data.select_dtypes(include='object').columns
        encoder = OneHotEncoder(drop='first', sparse=False)
        encoded_data = encoder.fit_transform(data[categorical_columns])
        encoded_df = pd.DataFrame(encoded_data, columns=encoder.get_feature_names(categorical_columns))
        data = pd.concat([data.drop(columns=categorical_columns), encoded_df], axis=1)
        return data

    def encode_categorical_pd_dummies(self, data):
        # Encode categorical columns using pd.get_dummies
        categorical_columns = data.select_dtypes(include='object').columns
        data = pd.get_dummies(data, columns=categorical_columns, prefix=categorical_columns, drop_first=True)
        return data

    def scale_numerical(self, data):
        # Scale numerical columns using StandardScaler
        numerical_columns = data.select_dtypes(include=np.number).columns
        scaler = StandardScaler()
        data[numerical_columns] = scaler.fit_transform(data[numerical_columns])
        return data

    def convert_datetime_columns(self, data, datetime_columns):
        # Convert datetime columns to datetime objects
        for col in datetime_columns:
            data[col] = pd.to_datetime(data[col], errors='coerce')
        return data


# Example usage:
if __name__ == "__main__":
    auto_preprocessor = AutoPreprocessor()  # No need to set imputation methods during initialization
    raw_data = pd.read_csv("puntonet_englist_org.csv")

    # Define a list of columns that contain datetime strings
    datetime_columns = ['date_for_last_update', 'completion_date',
                        'creation_date']  # Replace with your actual column names

    # Choose different imputation methods, categorical encoding methods, and encoding methods based on your dataset
    numerical_impute_method = 'mean'
    categorical_impute_method = 'knn'
    categorical_encoding_method = 'label'  # Choose 'label', 'one-hot', or 'pd-dummies'

    preprocessed_data = auto_preprocessor.preprocess(raw_data, datetime_columns=datetime_columns,
                                                     numerical_impute_method=numerical_impute_method,
                                                     categorical_impute_method=categorical_impute_method,
                                                     categorical_encoding_method=categorical_encoding_method)

    print(preprocessed_data.head())
